"""SNAPtoolbelt Mesh/Network Commands"""
from __future__ import print_function

import itertools
import logging
import os

import progressbar

from arghandler import ArgumentHandler
from argparse import RawDescriptionHelpFormatter
from textwrap import dedent
from tornado.ioloop import PeriodicCallback

from snap_profile.network import Network

import snapstraction.mesh
import snapstraction.node_finder
import snapstraction.topology

import snaptoolbelt.topology_utils

from snaptoolbelt.alias import Aliases
from snaptoolbelt.util import validate_mac_address, eprint, exit_with_usage

log = logging.getLogger(__name__)


def network(parser, context, args):
    """Network-related subcommands"""
    handler = ArgumentHandler(
        usage="snap [global options] network [-h] (dm-ping|intercept|listen|ping|topology) [options]",
        description=dedent("""
        SNAP Network Commands:
          dm-ping: Ping the network, use directed multicast RPC for replies
          intercept: Listen for intercepted node output
          listen: Listen for all traffic
          ping: Ping the network, use unicast RPC for replies
          topology: Acquire the topology of the network
        """),
        formatter_class=RawDescriptionHelpFormatter,
    )

    handler.set_subcommands({
        'dm-ping': dm_ping,
        'intercept': intercept_listen,
        'listen': listen,
        'ping': ping,
        'topology': topology,
    })

    handler.run(args, context_fxn=lambda x: context)


def ping(parser, context, args):
    """Ping the network"""
    handler = ArgumentHandler(
        usage="snap [global options] network ping [-h] [-c <count>] [-t <ttl>]",
        description="Ping the network looking for nearby nodes, use unicast RPC for replies",
    )

    handler.add_argument(
        "-c",
        "--count",
        default=3,
        type=int,
        help="# of packets to send (default is 3)",
    )

    handler.add_argument(
        "-t",
        "--ttl",
        default=5,
        type=int,
        help="Multicast TTL for pings (default is 5)",
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to network ping command: %r", rest)

    log.info("Ping the network")
    eprint('Unicast Ping:')
    widgets = [progressbar.FormatLabel('Found 0 ()'), ' ',
               progressbar.Percentage(), ' ', progressbar.Bar(), ' ',
               progressbar.Timer(), ' ', progressbar.AdaptiveETA()]

    progress = progressbar.ProgressBar(widgets=widgets, max_value=args.count)
    progress.start()

    def on_progress(seconds, found_count, last_addr):
        """Update progress bar"""
        widgets[0] = progressbar.FormatLabel('Found %d (%s)' %
                                             (found_count, last_addr))
        progress.update(seconds)

    def on_finished():
        """Finish progress bar"""
        progress.finish()

    with context.gus() as gus:
        result = gus.run_sync(args.count + 1,
                              snapstraction.mesh.unicast_ping,
                              count=args.count,  # We ping each second
                              ttl=args.ttl,
                              progress_cb=on_progress,
                              finished_cb=on_finished)
        context.output(result)


def dm_ping(parser, context, args):
    """Ping that uses dmcast for the response instead of unicast"""
    handler = ArgumentHandler(
        usage="snap [global options] network dm-ping [-h] [-c <count>] [-t <ttl>]",
        description="Ping the network looking for nearby nodes, use directed multicast for replies",
    )

    handler.add_argument(
        "-c",
        "--count",
        default=3,
        type=int,
        help="# of packets to send (default is 3)",
    )

    handler.add_argument(
        "-t",
        "--ttl",
        default=5,
        type=int,
        help="Multicast TTL for pings (default is 5)",
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to network dm-ping command: %r", rest)

    log.info("Ping the network")
    eprint('Directed Multicast Ping:')
    widgets = [progressbar.FormatLabel('Found 0 ()'), ' ',
               progressbar.Percentage(), ' ', progressbar.Bar(), ' ',
               progressbar.Timer(), ' ', progressbar.AdaptiveETA()]

    progress = progressbar.ProgressBar(widgets=widgets, max_value=args.count)
    progress.start()

    def on_progress(seconds, found_count, last_addr):
        """Update progress bar"""
        widgets[0] = progressbar.FormatLabel('Found %d (%s)' %
                                             (found_count, last_addr))
        progress.update(seconds)

    def on_finished():
        """Finish progress bar"""
        progress.finish()

    with context.gus() as gus:
        result = gus.run_sync(args.count + 1,
                              snapstraction.mesh.dmcast_ping,
                              count=args.count,  # We ping each second
                              ttl=args.ttl,
                              progress_cb=on_progress,
                              finished_cb=on_finished)
        context.output(result)


def intercept_listen(parser, context, args):
    """Listen for intercepted node output"""
    handler = ArgumentHandler(
        usage="snap [global options] network intercept [-h] [--timeout <seconds>]",
        description="Listen for intercepted node output",
    )

    handler.add_argument(
        '-t',
        '--timeout',
        type=int,
        default=None,
        help='How many seconds to listen (default: forever)',
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to network intercept command: %r", rest)

    log.info("Trying to listen for intercepted output")

    with context.gus() as gus:
        gus.run_sync(None,
                     snapstraction.mesh.listen_for_intercepted_output,
                     context.output,
                     _timeout=args.timeout)


def find(parser, context, args):
    """find the target node"""
    handler = ArgumentHandler(
        usage="snap [global options] find [-h] <target(s)> [--list <list-file>] [--move <network>]",
        description="Find the target node(s), optionally moving them to the specified network",
    )

    handler.add_argument(
        'targets',
        nargs='?',
        default=None,
        help="Node(s) to find",
    )

    handler.add_argument(
        '-l',
        '--list',
        dest="list_file",
        help="File containing list of nodes to find",
    )

    handler.add_argument(
        '-m',
        '--move',
        help="Profile to move nodes to.",
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to find command: %r", rest)

    with Aliases() as aliases:

        if args.list_file:
            # do expansion of list file name/path
            list_file_path = os.path.join(os.path.curdir, os.path.expanduser(args.list_file))
            targets = [aliases.translate(s) for s in snapstraction.node_finder.get_targets_from_file(list_file_path)]
        elif args.targets:
            targets = [aliases.translate(s.strip()) for s in args.targets.split(',')]
        else:
            log.error("Must provide either comma-separated target addresses (123456,234567,...) or a file with a list of targets, one per line")
            raise Exception("No targets provided")

    # load the network data to pass down
    move_to_network = Network.load(args.move) if args.move else None

    log.info("finding node(s), optionally moving to target network...")

    # validate the address format
    validate_mac_address(targets)

    message = find_nodes(context, targets, move_to_network)

    context.output(message)


def find_nodes(context, targets, move_to_network):
    """Load finder script to run along with progressbar.

    :param context: context is passed to get gus and other important params
    :param targets: list of all the targets user want to find.
    :param move_to_network: a network object to move the nodes.
    :returns result: returns a list of dictionary containing the node info.
    :returns final_message: returns the summary of nodes.
    """

    custom_pre_text = progressbar.FormatCustomText(
        'Searching %(address)s on channel %(channel)s',
        dict(
            address='', channel=''
        ),
    )

    widgets = [custom_pre_text, ' : ',
               progressbar.AnimatedMarker(),
               ]

    progress = progressbar.ProgressBar(widgets=widgets, max_value=16)
    progress.start()

    counter = itertools.cycle(xrange(16))

    def progress_up():
        progress.update(counter.next())

    p_cb = PeriodicCallback(progress_up, 250)

    def progress_cb(address, channel):
        log.debug("Got progress cb: %r %r", address, channel)

        custom_pre_text = progress.widgets[0]
        custom_pre_text.update_mapping(address=address, channel=channel)
        progress.start()
        p_cb.start()

    def finished_cb(address, data):
        log.debug("Got finished cb: %r %r", address, data)
        p_cb.stop()

        if not p_cb.is_running():
            context.output(data)

    with context.gus() as gus:
        result = gus.run_sync(context.timeout,
                              snapstraction.node_finder.finder,
                              targets, move_to_network=move_to_network,
                              progress_cb=progress_cb,
                              finished_cb=finished_cb)

    total = len(targets)
    found = 0

    if result:
        for key in result:
            if result[key]['on_channel']:
                found = found + 1

    final_message = "Done. {0}/{1} Found".format(found, total)
    return final_message


def topology(parser, context, args):
    """Take the topology of the network"""

    handler = ArgumentHandler(
        usage="snap [global options] network topology [-h] <root node> [-d [-c]] [-s]",
        description="Gather the topology of the network, starting at the root (defaults to the bridge)",
    )

    handler.add_argument(
        "root_node",
        nargs="?",
        default="bridge",
        help="Root of topology (starting point)",
    )

    handler.add_argument(
        "-d",
        "--dot",
        help="Output DOT digraph format",
        action="store_true",
    )

    handler.add_argument(
        "-c",
        "--colorize-lq",
        help="Colorize edges based on link quality. (>-30 dBm is excellent, <-80 dBm is bad.)",
        action="store_true",
    )

    handler.add_argument(
        "-s",
        "--simplify",
        help="Merge digraph into undirected graph format",
        action="store_true",
    )

    args, rest = handler.parse_known_args(args)
    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to network topology command: %r", rest)

    validate_mac_address(args.root_node)

    log.info("Get the topology of the network")

    with context.gus() as gus:
        result = gus.run_sync(
            context.timeout,
            snapstraction.topology.take_network_census,
            args.root_node,
        )
        if args.simplify:
            result = snaptoolbelt.topology_utils.simplify_topology(result)

        if args.dot:
            if args.simplify:
                func = snaptoolbelt.topology_utils.render_as_dot_simple_graph
            else:
                func = snaptoolbelt.topology_utils.render_as_dot_digraph

            for line in func(result, colorize_lq=args.colorize_lq):
                print(line)
        else:
            context.output(result['topology'])


def listen(parser, context, args):
    """Listen to packets received"""
    handler = ArgumentHandler(
        usage="snap [global options] network listen [-h] [--timeout <seconds>]",
        description="Listen for all traffic",
    )

    handler.add_argument(
        '-t',
        '--timeout',
        type=int,
        default=None,
        help='How many seconds to listen (default: forever)'
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to network listen command: %r", rest)

    log.info("Listening to the network")

    with context.gus() as gus:
        gus.run_sync(None,
                     snapstraction.mesh.listen,
                     context.output,
                     _timeout=args.timeout)
