# -*- coding: utf-8 -*-
"""SNAPtoolbelt Node commands"""
from __future__ import print_function

import logging
import os
import sys
import traceback

from argparse import RawDescriptionHelpFormatter, SUPPRESS
from textwrap import dedent

import progressbar
import serialwrapper

from arghandler import ArgumentHandler

from snap_profile.network import Network

import snapstraction.gus
import snapstraction.build
import snapstraction.node
import snapstraction.nvparam
import snapstraction.node_finder
import snapstraction.topology

import snaptoolbelt.topology_utils

from snap_profile.profile import Profile

from snapstraction.node_cache import NodeCache
from snapstraction.util import parse_value_for_snap, validate_snap_address, clean_up

from snaptoolbelt.alias import Aliases
from snaptoolbelt.util import mkdir_p, validate_package, eprint, exit_with_error, exit_with_usage

log = logging.getLogger(__name__)


class IncompleteResultError(Exception):
    """Raise this when you get a partial response"""
    pass


def update_cache(context, gus):
    """Add the context's target address to the node cache"""
    # Update the context's target so we record the bridge address correctly
    context.target = clean_up(context.target, gus.bridge_address)

    with NodeCache() as cache:
        cache.add(context.target)


def node(parser, context, args):
    """Node-related subcommands"""
    subcommands = {
        'energy': energy,
        'firmware': firmware,
        'info': info,
        'intercept': intercept,
        'move': move,
        'nvparam': nvparam,
        'ping': ping,
        'reboot': reboot,
        'script': script,
        'topology': topology,
        'traceroute': traceroute,
    }

    handler = ArgumentHandler(
        usage="snap [global options] node <target> [-h] <command> [options]",
        description=dedent("""
        SNAPToolbelt Node Commands
        --------------------------

        * energy: Ask the node to perform a clear channel scan
        * firmware: SNAP firmware commands
        * info: Query node information
        * intercept: Ask the node to start sending us output
        * move: Move the node to a network
        * nvparam: Query / Set NVParams on the node
        * ping: Ping the node (get the link quality)
        * reboot: Ask the node to reboot
        * script: SNAPpy script commands
        * topology: Request topology information from the node
        * traceroute: Perform a traceroute to the node
        """),
        formatter_class=RawDescriptionHelpFormatter,
    )

    # All node commands take a single target address
    handler.add_argument(
        "target",
        help="6-char node address, or 'bridge' for the bridge node")

    # Add subcommands here. Again, the @subcmd decorator doesn't
    # quite do the right thing for us, so we have to manually register.
    handler.set_subcommands(subcommands)

    # The context_fxn passes the global context down to the
    # subcommands so they can get to things like the profile's
    # crypto settings

    def ctx(node_context):
        """Update the context for node commands"""
        # Translate aliases
        with Aliases() as aliases:
            node_context.target = aliases.translate(node_context.target)

        # Check the address
        address = validate_snap_address(node_context.target)

        if address is None:
            raise Exception("%s is not a valid 6 hex char address" %
                            node_context.target)

        # Update the context
        node_context.target = address

        # Merge the parent context with the node context, using vars to get dicts
        vars(context).update(vars(node_context))

        return context

    try:
        handler.run(args, context_fxn=ctx)
    except serialwrapper.PyserialDriver.SerialOpenException:
        exit_with_error("Could not open serial port: {0}".format(getattr(Profile.load('default'), 'device')))
    except snapstraction.gus.NoResponseError:
        exit_with_error("\nUnable to send RPC to {}".format(context.target))
    except Exception as ex:
        # If debug logging is specified, we still want to see tracebacks
        if context.log_level == logging.DEBUG:
            traceback.print_exc()
        exit_with_error(ex)


def info(parser, context, args):  # noqa
    """Get info from the node at the given address"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> info [-h] (all|device|firmware|mcast|mesh|network|script|security|stats|uart)",
        description="Query information for the target node.",
        add_help=False,
    )

    handler.add_argument("subcommand", nargs='?', default='summary',
                         help="info category ('all' for everything, or blank for summary info)")

    args, rest = handler.parse_known_args(args)

    if '-h' in rest or '--help' in rest:
        handler.print_help()
        sys.exit(0)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to info command: %r", rest)

    log.info("Getting %s info from %s",
             args.subcommand, context.target)

    info_categories = {
        'device': snapstraction.node.device_info,
        'firmware': snapstraction.node.firmware_info,
        'mcast': snapstraction.node.mcast_info,
        'mesh': snapstraction.node.mesh_info,
        'network': snapstraction.node.network_info,
        'script': snapstraction.node.script_info,
        'security': snapstraction.node.security_info,
        'stats': snapstraction.node.stats_info,
        'summary': snapstraction.node.summary_info,
        'uart': snapstraction.node.uart_info,
    }

    if args.subcommand not in info_categories and args.subcommand != 'all':
        handler.print_help()
        sys.exit(1)

    try:
        data = {}
        with context.gus() as gus:
            # Try to ping first, if that doesn't work, abort!
            try:
                gus.run_sync(context.timeout, snapstraction.node.ping, context.target)
            except snapstraction.gus.NoResponseError:
                log.error("Unable to ping node, aborting!")
                raise Exception("Unable to ping node, aborting!")

            if args.subcommand == 'all':
                errs = 0
                for category in sorted(info_categories.keys()):
                    info_func = info_categories[category]
                    result = gus.run_sync(context.timeout, info_func, context.target)
                    errs += result['_errors']
                    data.update(**result)
                data['_errors'] = errs
            else:
                data = gus.run_sync(context.timeout, info_categories[args.subcommand], context.target)

            update_cache(context, gus)
    except:
        raise
    else:
        # Now we need to merge all the results
        errors = data.get('_errors', 0)
        del data['_errors']

        context.output(data)

        if errors:
            raise IncompleteResultError()


def traceroute(parser, context, args):
    """Perform a traceroute on the target node"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> traceroute [-h]",
        description="Traceroute to the target node.",
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node traceroute command: %r", rest)

    log.info("Trying to perform traceroute on node %s", context.target)
    with context.gus() as gus:
        result = gus.run_sync(context.timeout, snapstraction.node.traceroute,
                              context.target)
        update_cache(context, gus)
        context.output(result)


def nvparam(parser, context, args):
    """NVParam Git-alike subcommand"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> nvparam [-h] (all | <nvparam> [value])",
        description=dedent("""
            Query/Set Node NVParams:
              <nvparam>: Query NVParam by id
              all: Query NVParams 0-127 (all non-user params)
              <nvparam> <value>: Set NVParam id to value
        """),
        formatter_class=RawDescriptionHelpFormatter,
    )

    handler.add_argument("nvparam", help="ID of NVParam (integer or name)")
    handler.add_argument("value", nargs='?', default=None, help="Value to set")

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node nvparam command: %r", rest)

    log.debug(args)

    if args.value is None:
        if args.nvparam == "all":
            result = {}
            with context.gus() as gus:
                for i in xrange(128):
                    value = gus.run_sync(context.timeout,
                                         snapstraction.nvparam.get_nvparam,
                                         context.target, "%d" % i)
                    for k, v in value.iteritems():
                        result.update({k: v})
                update_cache(context, gus)
            context.output(result)

        else:
            log.info("Reading NVParam %s", args.nvparam)
            with context.gus() as gus:
                result = gus.run_sync(context.timeout,
                                      snapstraction.nvparam.get_nvparam,
                                      context.target, args.nvparam)
                update_cache(context, gus)
                context.output(result)
    else:
        args.value = parse_value_for_snap(args.value)
        log.info("Setting NVparam %s to %r", args.nvparam, args.value)
        with context.gus() as gus:
            result = gus.run_sync(
                context.timeout, snapstraction.nvparam.set_nvparam,
                context.target, args.nvparam, args.value)
            update_cache(context, gus)
            context.output(result)


def reboot(parser, context, args):
    """Reboot the target node"""
    handler = ArgumentHandler(
        usage="snap [global options] node target reboot [-h] [delay]",
        description="Reboot the target node, with the optional delay"
    )

    handler.add_argument('delay',
                         default=None,
                         type=int,
                         nargs='?',
                         help="Optional ms to delay before rebooting.")
    log.debug("ARGS: %r", args)

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node reboot command: %r", rest)

    log.info("Trying to reboot node %s, args: %r", context.target,
             args)

    with context.gus() as gus:
        result = gus.run_sync(context.timeout,
                              snapstraction.node.reboot,
                              context.target,
                              delay=args.delay)
        update_cache(context, gus)
        context.output(result)


def script(parser, context, args):
    """Script Commands"""

    log.debug("script context: %r", context)
    log.debug("script args: %r", args)

    handler = ArgumentHandler(
        usage="snap [global options] node <target> script [-h] (build|erase|info|upload) [options]",
        description=dedent("""
        SNAPpy Script Commands:
          build: Build script for this node
          erase: Erase the script on this node
          info: Query currently loaded script information
          upload: Build script and upload image to this node
        """),
        formatter_class=RawDescriptionHelpFormatter,
        add_help=False,
    )

    handler.add_argument("subcommand", nargs='?', default=None,
                         help="script subcommand (or blank for info)")

    args, rest = handler.parse_known_args(args)

    if not args.subcommand and '-h' not in rest and '--help' not in rest:
        # Just get info
        script_info(parser, context, rest)
    else:
        if args.subcommand == "build":
            build(parser, context, rest)
        elif args.subcommand == "upload":
            upload(parser, context, rest)
        elif args.subcommand == "erase":
            script_erase(parser, context, rest)
        elif args.subcommand == "info":
            script_info(parser, context, rest)
        else:
            if '-h' in rest or '--help' in rest:
                handler.print_help()
                sys.exit(0)
            exit_with_usage("Unexpected command: %r", args.subcommand)


def script_info(parser, context, args):
    """Get script-related info from the node"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> script info [-h]",
        description="Query the currently loaded script on the node.",

    )
    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node script info command: %r", rest)

    log.info("Getting script info from node %s", context.target)

    with context.gus() as gus:
        result = gus.run_sync(context.timeout, snapstraction.node.script_info,
                              context.target)
        update_cache(context, gus)
        del result['_errors']
        context.output(result)


def build(parser, context, args):  # noqa
    """Build the SCRIPT for the NODE"""
    # Verify snapbuild installed in the environment or not
    validate_package('snapbuild')
    handler = ArgumentHandler(
        usage="snap [global options] node <target> script build <script> [-h] [-I include [...]] [-o output]",
        description="Build the script for the target node.",
    )

    handler.add_argument('script', help="SNAPpy script to build")
    handler.add_argument(
        '-I',
        '--Include',
        dest='include',
        default=[],
        metavar='IMPORT_PATH',
        help='Allow imports from additional path',
        action='append',
    )
    handler.add_argument(
        "-o",
        "--output",
        dest="output",
        metavar="SPY",
        help=dedent("""
            Path of SNAPpy script image (and map file), defaults to ./SCRIPT.spy and ./SCRIPT.map.
            If this doesn't already end in '.spy', '.spy' will be appended.
        """),
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node script build command: %r", rest)

    data = {}

    args.script = os.path.abspath(os.path.expanduser(args.script))

    # Need to build it.
    # First, find out what module/core to use
    with context.gus() as gus:
        result = gus.run_sync(context.timeout,
                              snapstraction.node.get_node_core_module,
                              context.target)

        core = result.get('core', None)
        module = result.get('module', None)
        platform = result.get('platform', None)

        log.info("Node %s is a(n) %s running Core %s, Platform (NV41) %s",
                 context.target, module, core, platform)

        data['core'] = core
        data['module'] = module
        data['platform'] = platform

        update_cache(context, gus)

    if not core or not module:
        log.error("Could not determine target's core and module")
        raise Exception("Could not determine target's core and module")

    data['includes'] = args.include

    if args.output:
        output_path = args.output
    else:
        output_path = os.path.join(
            '.', os.path.splitext(os.path.basename(args.script))[0])

    data['output_dir'] = os.path.dirname(output_path)

    crc = snapstraction.build.call(args.script,
                                   module,
                                   core,
                                   includes=args.include,
                                   platform=platform,
                                   output_path=output_path)

    data['crc'] = crc

    spy_out = os.path.abspath(os.path.splitext(output_path)[0] + '.spy')
    data['spy'] = spy_out
    data['map'] = os.path.abspath(os.path.splitext(output_path)[0] + '.map')

    context.output(data)


def upload(parser, context, args):  # noqa
    """Upload the SCRIPT to the NODE"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> script upload <script.[s]py> [-h] [-I include [...]]",
        description=dedent("""
            Upload the script to the target node.
            If a SNAPpy script (.py) is specified, build the script first, then upload.
            If a script image (.spy) is specified, just upload it.
        """),
    )

    handler.add_argument('script', help="SNAPpy script to upload")
    handler.add_argument('-I',
                         '--Include',
                         dest='include',
                         default=[],
                         metavar='IMPORT_PATH',
                         help='Allow imports from additional path',
                         action='append')
    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node script upload command: %r", rest)

    data = {}

    args.script = os.path.abspath(os.path.expanduser(args.script))

    # If the script is a SPY file, just load it.
    if os.path.splitext(args.script)[1].lower() == '.spy':
        data['spy'] = args.script
        result = ota_script_upload(context, args.script)
        data.update(**result)
        context.output(data)

    elif os.path.splitext(args.script)[1].lower() == '.py':
        # Need to build it.

        # Verify snapbuild installed in the environment
        validate_package('snapbuild')

        # First, find out what module/core to use
        with context.gus() as gus:
            result = gus.run_sync(context.timeout,
                                  snapstraction.node.get_node_core_module,
                                  context.target)

            core = result.get('core', None)
            module = result.get('module', None)
            platform = result.get('platform', None)

            log.info("Node %s is a(n) %s running Core %s, Platform (NV41) %s",
                     context.target, module, core, platform)

            update_cache(context, gus)

            data['core'] = core
            data['module'] = module
            data['platform'] = platform

        if not core or not module:
            log.error("Could not determine target's core and module")
            raise Exception("Could not determine target's core and module")

        data['includes'] = args.include

        spycache = os.path.expanduser(os.path.join('~', '.snap', 'spycache'))
        data['output_dir'] = spycache
        mkdir_p(spycache)

        output_path = os.path.join(
            spycache, os.path.splitext(os.path.basename(args.script))[0])

        crc = snapstraction.build.call(args.script,
                                       module,
                                       core,
                                       includes=args.include,
                                       platform=platform,
                                       output_path=output_path)

        data['crc'] = crc

        spy_out = os.path.abspath(os.path.splitext(output_path)[0] + '.spy')
        data['spy'] = spy_out
        data['map'] = os.path.abspath(os.path.splitext(output_path)[0] +
                                      '.map')
        result = ota_script_upload(context, spy_out)
        data.update(**result)
        context.output(data)

    upload_status = data.get('status_code', -1)
    if upload_status:
        exit_with_error("Error uploading script: %s" % data.get('message', 'Unknown: %d' % upload_status),
                        exit_code=upload_status)


def script_erase(parser, context, args):
    """Erase the script from the node"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> script erase [-h]",
        description="Erase the script on the target node.",
    )
    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node script erase command: %r", rest)

    log.info("Erasing the script from node %s", context.target)

    with context.gus() as gus:
        result = gus.run_sync(context.timeout, snapstraction.node.script_erase,
                              context.target)
        update_cache(context, gus)
        context.output(result)


def ota_script_upload(context, spy_file):
    """Load script OTA"""

    spy_path = os.path.join(
        os.path.curdir, os.path.expanduser(spy_file)) if spy_file else None

    eprint('OTA Script Upload: %s ==> %s' % (os.path.basename(spy_path), context.target))
    widgets = [progressbar.FormatLabel('Uploading:'), ' ',
               progressbar.Percentage(), ' ', progressbar.Bar(), ' ',
               progressbar.Timer(), ' ', progressbar.AdaptiveETA(), ' ',
               progressbar.AdaptiveTransferSpeed()]

    progress = progressbar.ProgressBar(widgets=widgets, max_value=100)
    progress.start()

    def on_progress(addr, percent_complete):
        """Update progress bar"""
        progress.update(min(percent_complete, 100))

    def on_finished(addr, status):
        """Finish progress bar"""
        progress.finish()

    with context.gus() as gus:
        if context.unicast:
            result = gus.run_sync(context.timeout,
                                  snapstraction.node.unicast_ota_script_upload,
                                  context.target,
                                  filename=spy_path,
                                  progress_cb=on_progress,
                                  finished_cb=on_finished)
        elif context.dmcast:
            result = gus.run_sync(context.timeout,
                                  snapstraction.node.dmcast_ota_script_upload,
                                  context.target,
                                  filename=spy_path,
                                  progress_cb=on_progress,
                                  finished_cb=on_finished)
        else:
            result = gus.run_sync(context.timeout,
                                  snapstraction.node.ota_script_upload,
                                  context.target,
                                  filename=spy_path,
                                  progress_cb=on_progress,
                                  finished_cb=on_finished)
        update_cache(context, gus)
        # The ota_script_upload from SNAPstraction now always returns a list of a single result
        return result[0]


def firmware(parser, context, args):
    """Firmware Commands"""

    log.debug("firmware context: %r", context)
    log.debug("firmware args: %r", args)

    handler = ArgumentHandler(
        usage="snap [global options] node <target> firmware [-h] (info | upload) [options]",
        description=dedent("""
        OTA firmware commands:
          info: Query firmware information
          upload: Load firmware OTA
        """),
        formatter_class=RawDescriptionHelpFormatter,
        add_help=False,
    )

    handler.add_argument("subcommand", nargs='?', default=None,
                         help="firmware subcommand (or blank for info)")

    args, rest = handler.parse_known_args(args)

    if not args.subcommand and '-h' not in rest and '--help' not in rest:
        # Just get info
        firmware_info(parser, context, rest)
    else:
        if args.subcommand == "info":
            firmware_info(parser, context, rest)
        elif args.subcommand == "upload":
            ota_firmware_load(parser, context, rest)
        else:
            if '-h' in rest or '--help' in rest:
                handler.print_help()
                sys.exit(0)

            exit_with_usage("Unexpected command: %r", args.subcommand)


def firmware_info(parser, context, args):
    """Get firmware-related info from the node"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> firmware info [-h]",
        description="Query the target node's firmware version",
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node firmware info command: %r", rest)

    log.info("Getting firmware info from node %s", context.target)

    with context.gus() as gus:
        result = gus.run_sync(context.timeout,
                              snapstraction.node.firmware_info,
                              context.target)
        update_cache(context, gus)
        del result['_errors']
        context.output(result)


def ota_firmware_load(parser, context, args):  # noqa - code complexity
    """OTA firmware upgrade"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> firmware upload [-h] <core>",
        description="Load the specified firmware on the target node OTA.",
    )

    handler.add_argument('core',
                         help="Core firmware version (e.g. 2.7.2)")

    # Hidden option
    handler.add_argument('custom_filename',
                         default=None,
                         nargs='?',
                         help=SUPPRESS)

    args, rest = handler.parse_known_args(args)

    custom_fullpath = None

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node firmware load command: %r", rest)

    if args.core == 'custom':
        if args.custom_filename is None:
            exit_with_usage(handler, log, "No custom firmware specified.")

        custom_fullpath = os.path.join(os.path.curdir, os.path.expanduser(args.custom_filename))
        if not os.path.exists(custom_fullpath):
            exit_with_usage(handler, log, "Could not locate custom firmware: %r", custom_fullpath)

    eprint("The target node may be unresponsive for up to 2 minutes *after* the upload completes.")
    eprint("Do not remove power from the node during this time.")

    eprint('OTA Firmware Load: %s ==> %s:' % (args.core, context.target))

    widgets = [progressbar.FormatLabel('Uploading:'), ' ',
               progressbar.Percentage(), ' ', progressbar.Bar(), ' ',
               progressbar.Timer(), ' ', progressbar.AdaptiveETA(), ' ',
               progressbar.AdaptiveTransferSpeed()]

    progress = progressbar.ProgressBar(widgets=widgets, max_value=100)
    progress.start()

    def on_progress(addr, percent_complete):
        """Update progress bar"""
        pct = min(percent_complete, 100)
        progress.update(pct)
        if pct == 100:
            eprint("Pinging node to verify if upgrade process is complete")

    def on_finished(addr, status):
        """Finish progress bar"""
        progress.finish()

    try:
        with context.gus() as gus:
            # NB: Custom OTA loads are *always* unicast.
            if custom_fullpath is not None:
                result = gus.run_sync(context.timeout,
                                      snapstraction.node.unicast_ota_firmware_upload,
                                      context.target,
                                      None,
                                      custom_filename=custom_fullpath,
                                      progress_cb=on_progress,
                                      finished_cb=on_finished)
            else:
                if context.unicast:
                    result = gus.run_sync(context.timeout,
                                          snapstraction.node.unicast_ota_firmware_upload,
                                          context.target,
                                          args.core,
                                          progress_cb=on_progress,
                                          finished_cb=on_finished)
                elif context.dmcast:
                    result = gus.run_sync(context.timeout,
                                          snapstraction.node.dmcast_ota_firmware_upload,
                                          context.target,
                                          args.core,
                                          progress_cb=on_progress,
                                          finished_cb=on_finished)
                else:
                    result = gus.run_sync(context.timeout,
                                          snapstraction.node.ota_firmware_upload,
                                          context.target,
                                          args.core,
                                          progress_cb=on_progress,
                                          finished_cb=on_finished)
            # Unpack from list of single result
            result = result[0]
            update_cache(context, gus)
            context.output(result)
    except ImportError:
        log.error("Couldn't find firmware support package for core '%s', you may need to install it " +
                  "by running the following command:\n" +
                  "pip install --extra-index-url https://update.synapse-wireless.com/pypi snap-firmware-%s",
                  args.core, args.core)
        message = ("Couldn't find firmware support package for core '%s', you may need to install it "
                   "by running the following command:\n"
                   "pip install --extra-index-url https://update.synapse-wireless.com/pypi snap-firmware-%s")
        log.debug(message % (args.core, args.core))
        exit_with_error(message % (args.core, args.core))
    except Exception as ex:
        # If debug logging is specified, we still want to see tracebacks
        if context.log_level == logging.DEBUG:
            traceback.print_exc()
        exit_with_error(ex)
    else:
        upload_status = result.get('status_code', -1)
        if upload_status:
            exit_with_error("Error uploading firmware: %s" % result.get('message', 'Unknown: %d' % upload_status),
                            exit_code=upload_status)


def energy(parser, context, args):
    """Do an RF energy scan"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> energy [-h] [--num-queries N] [--delay SECONDS]",
        description="Ask the node to perform one or more energy scans",
    )

    handler.add_argument(
        "-n",
        "--num-queries",
        default=1,
        type=int,
        help="Number of energy queries to average",
    )

    handler.add_argument(
        "-d",
        "--delay",
        default=0,
        type=float,
        help="delay in seconds between queries",
    )

    handler.add_argument(
        "-c",
        "--cutoff",
        default=90,
        type=int,
        help="cutoff value for the dBm floor")

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node energy command: %r", rest)

    log.info("Trying to scanEnergy from node %s", context.target)

    with context.gus() as gus:
        result = gus.run_sync(context.timeout,
                              snapstraction.node.percent_clear_energies,
                              context.target,
                              num=args.num_queries,
                              delay=args.delay,
                              cutoff=args.cutoff)

        update_cache(context, gus)
    context.output(result)


def intercept(parser, context, args):
    """Node output intercept commands"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> intercept [-h] (start | stop) [options]",
        description="Control STDOUT/STDERR intercepting on the target node.",
    )

    handler.set_subcommands({
        'start': intercept_start,
        'stop': intercept_stop
    })

    handler.run(args, context_fxn=lambda x: context)


def intercept_start(parser, context, args):
    """Tell the node to start sending us its output"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> intercept start [-h] [--out-only | --err-only]",
        description="Start intercepting STDOUT/STDERR on the target node.",
    )

    sources_group = handler.add_mutually_exclusive_group()
    sources_group.add_argument('--out-only', action='store_true')
    sources_group.add_argument('--err-only', action='store_true')

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node intercept start command: %r", rest)

    intercept_stdout = not args.err_only
    intercept_stderr = not args.out_only

    sources = [('stdout', intercept_stdout), ('stderr', intercept_stderr)]
    log.info("Trying to start intercepting %s from node %s",
             ' and '.join([name for (name, enabled) in sources if enabled]),
             context.target)

    with context.gus() as gus:
        result = gus.run_sync(
            context.timeout, snapstraction.node.start_output_interceptor,
            context.target, intercept_stdout, intercept_stderr)
        update_cache(context, gus)
        context.output(result)


def intercept_stop(parser, context, args):
    """Tell the node to stop sending us output"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> intercept stop [-h]",
        description="Stop intercepting STDOUT/STDERR on the target node.",
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        exit_with_usage(handler, log, "Unexpected arguments to node intercept stop command: %r", rest)

    sources = ['stdout', 'stderr']
    log.info("Trying to stop intercepting %s from node %s",
             ' and '.join(sources), context.target)
    with context.gus() as gus:
        result = gus.run_sync(context.timeout,
                              snapstraction.node.stop_output_interceptor,
                              context.target)
        update_cache(context, gus)
        context.output(result)


def ping(parser, context, args):
    """Unicast ping to a single node"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> ping [-h]",
        description="Ping the target node with unicast RPCs.",
    )

    args = handler.parse_args(args)

    log.info("Pinging node %s...", context.target)
    with context.gus() as gus:
        result = gus.run_sync(context.timeout, snapstraction.node.ping,
                              context.target)
        update_cache(context, gus)
        context.output(result)


def move(parser, context, args):
    """
Move the target node to the target network.

Assumes you can already talk to the target.
If the target is the bridge, then update the profile's "last used" network.
    """

    handler = ArgumentHandler(
        usage="snap [global options] node <target> move [-h] <network>",
        description="Move the node to the target network.",
    )

    handler.add_argument('network', help="Network to move the node to")
    args = handler.parse_args(args)

    log.info("Moving node %s to %s...", context.target,
             args.network)

    try:
        target_network = Network.load(args.network)
    except:  # pylint: disable=W0702
        log.error("Unable to load target network %s", args.network)
        raise Exception("Unable to load target network {}".format(args.network))

    with context.gus() as gus:
        context.target = clean_up(context.target,
                                  gus.bridge_address)
        result = gus.run_sync(context.timeout,
                              snapstraction.node_finder.move_target_to_network,
                              context.target, target_network)

        if context.target in (gus.bridge_address, 'bridge'):
            # Update the profile's network to where we just moved the bridge node
            context.profile.network = args.network
            context.profile.save()

        update_cache(context, gus)
        context.output(result)


def topology(parser, context, args):
    """Request topology from a single node"""
    handler = ArgumentHandler(
        usage="snap [global options] node <target> topology [-h] [-d [-c]] [-s]",
        description="Ask the target node to acquire its local topology.",
    )

    handler.add_argument(
        "-d",
        "--dot",
        help="Output DOT digraph format",
        action="store_true",
    )

    handler.add_argument(
        "-c",
        "--colorize-lq",
        help="Colorize edges based on link quality. (>-30 dBm is excellent, <-80 dBm is bad.)",
        action="store_true",
    )

    handler.add_argument(
        "-s",
        "--simplify",
        help="Merge digraph into undirected graph format",
        action="store_true",
    )

    args = handler.parse_args(args)

    log.info("Acquiring topology from node %s...", context.target)
    with context.gus() as gus:
        result = gus.run_sync(
            context.timeout,
            snapstraction.topology.take_node_census,
            context.target,
        )
        update_cache(context, gus)
        if args.simplify:
            result = snaptoolbelt.topology_utils.simplify_topology(result)

        if args.dot:
            if args.simplify:
                func = snaptoolbelt.topology_utils.render_as_dot_simple_graph
            else:
                func = snaptoolbelt.topology_utils.render_as_dot_digraph

            for line in func(result, colorize_lq=args.colorize_lq):
                print(line)
        else:
            context.output(result['topology'])
