"""SNAPtoolbelt Recover Flash Commands"""
import logging
import os
import sys
from textwrap import dedent
from argparse import RawDescriptionHelpFormatter
import serialwrapper
from arghandler import ArgumentHandler
import progressbar

from snap_profile.profile import Profile
import snapstraction.flash

from snaptoolbelt.util import eprint, validate_package, exit_with_error, validate_SNAPcore_version, validate_SNAP_module
from snap_modules.snap_modules import get_parts, is_supported_combination, get_supported_cores_for_part, \
    UnknownCoreError, UnknownPartError, get_cores

log = logging.getLogger(__name__)

RF100_MODULES = ['RF100']


def _eprint_reset_message(module):
    if module in RF100_MODULES:
        eprint("Reset your device. RF100s currently don't reset-on-DTR, you must reset them manually.")
    else:
        eprint("Reset your device. We'll try to reset it for you if it's an SS200 or SN220.")


def recover(parser, context, args):
    """SNAP Recovery commands"""
    handler = ArgumentHandler(
        usage="snap [global options] recover [-h] (default-nvparam|erase-script|firmware|sniffer) [options]",
        description=dedent("""
        SNAP Recovery Commands
          default-nvparam: Default NV Params on the serial bridge node
          erase-script: Erase the script on the serial bridge node
          firmware: Flash SNAP Firmware on the serial bridge node
          sniffer: Flash SNAP Sniffer Firmware on the serial bridge node
        """),
        formatter_class=RawDescriptionHelpFormatter,
    )

    # Add subcommands here.
    handler.set_subcommands({
        'default-nvparam': default_nv,
        'erase-script': erase,
        'firmware': firmware,
        'sniffer': sniffer,
    })

    if context.profile.kind != 'serial':
        log.error(
            "Recovery commands only work with direct serial access, %s not supported.",
            context.profile['snap_bridge_type'])
        sys.exit(1)

    handler.run(args, context_fxn=lambda x: context)


def firmware(parser, context, args):  # noqa
    """Upload firmware via recovery"""
    handler = ArgumentHandler(
        usage="snap [global options] recover firmware [-h] <module> (<core> | -f <file>)",
        description="Flash the SNAP firmware on the bridge node",
    )

    handler.add_argument('module',
                         help="Module name (i.e. 'RF220UF1', 'SS200')")
    handler.add_argument('core',
                         default=None,
                         nargs='?',
                         help="SNAP Core version (i.e. '2.7.2')")
    handler.add_argument('-f', '--file', default=None, help="Override firmware file path")

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to recover firmware command: %r", rest)
        sys.exit(1)

    if args.core is None and args.file is None:
        log.error("Must specify either core version or override path.")
        sys.exit(1)

    validate_SNAP_module(args.module)

    if args.core and not args.file:

        validate_SNAPcore_version(args.core)

        try:
            if not is_supported_combination(args.module, args.core):
                exit_with_error("Unknown SNAPcore version: %s. Valid SNAPcore versions for SNAP module %s are: %s"
                                % (args.core, args.module, ', '.join(get_supported_cores_for_part(args.module))))
        except UnknownCoreError:
            exit_with_error("Unknown SNAPcore version: %s" % args.core,
                            "Valid SNAPcore versions are %s" % ', '.join(get_cores()))
        except UnknownPartError:
            exit_with_error("Unknown SNAP module: %s" % args.module, "Valid SNAP modules are %s" % ', '.join(get_parts()))

        snap_firmware = "snap-firmware-%s" % args.core
        validate_package(snap_firmware)

    if args.file:
        if not os.path.exists(args.file) and not args.file.endswith(".sfi"):
            exit_with_error("Cannot access: %s. No such file or directory" % args.file)

    log.info("Loading firmware for module %s, core %s", args.module, args.core)

    _eprint_reset_message(args.module)

    eprint('Flashing Firmware:')
    widgets = [progressbar.FormatLabel('Processing:'), ' ',
               progressbar.Percentage(), ' ', progressbar.Bar(), ' ',
               progressbar.Timer(), ' ', progressbar.AdaptiveETA()]

    progress = progressbar.ProgressBar(widgets=widgets, max_value=1000)

    def on_progress(counter, new_max):
        """Update Progress Bar"""
        progress.max_value = max(new_max, counter)
        progress.update(counter)

    def on_finished():
        """Finish Progress Bar"""
        progress.finish()

    progress.start()

    try:
        snapstraction.flash.firmware(context.profile, args.module, args.core, sniffer=False, override_file=args.file,
                                     progress_cb=on_progress, finished_cb=on_finished)
    except serialwrapper.PyserialDriver.SerialOpenException:
        exit_with_error("Could not open serial port: {0}" .format(getattr(Profile.load('default'), 'device')))
    except snapstraction.flash.FlasherError:
        exit_with_error("Flasher Error: data timeout")


def sniffer(parser, context, args):  # noqa
    """Upload firmware via recovery"""
    handler = ArgumentHandler(
        usage="snap [global options] recover sniffer [-h] <module> [--file <file>]",
        description="Flash the Sniffer firmware on the bridge node",
    )

    handler.add_argument('module',
                         help="Module name (i.e. 'RF220UF1', 'SS200')")
    handler.add_argument('-f', '--file', default=None, help="Override firmware file path")

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to recover sniffer command: %r", rest)
        sys.exit(1)

    validate_SNAP_module(args.module)
    if args.module and not args.file:
        validate_package("snap-sniffer")

    if args.file:
        if not os.path.exists(args.file) and not args.file.endswith(".sfi"):
            exit_with_error("Cannot access: %s. No such file or directory" % args.file)

    log.info("Loading sniffer firmware for module %s", args.module)

    _eprint_reset_message(args.module)

    eprint('Flashing Sniffer:')
    widgets = [progressbar.FormatLabel('Processing:'), ' ',
               progressbar.Percentage(), ' ', progressbar.Bar(), ' ',
               progressbar.Timer(), ' ', progressbar.AdaptiveETA()]

    progress = progressbar.ProgressBar(widgets=widgets, max_value=1000)

    def on_progress(counter, new_max):
        """Update Progress Bar"""
        progress.max_value = max(new_max, counter)
        progress.update(counter)

    def on_finished():
        """Finish Progress Bar"""
        progress.finish()

    progress.start()

    try:
        snapstraction.flash.firmware(context.profile, args.module, '_', sniffer=True, override_file=args.file,
                                     progress_cb=on_progress, finished_cb=on_finished)
    except serialwrapper.PyserialDriver.SerialOpenException:
        exit_with_error("Could not open serial port: {0}".format(getattr(Profile.load('default'), 'device')))
    except snapstraction.flash.FlasherError:
        exit_with_error("Flasher Error: data timeout")


def default_nv(parser, context, args):
    """Reset NV Params via recovery"""
    handler = ArgumentHandler(
        usage="snap [global options] recover default-nvparam [-h] <module>",
        description="Default the NV Parameters on the bridge node",
    )

    handler.add_argument('module',
                         help="Module name (i.e. 'RF220UF1', 'SS200')")

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to recover default-nvparam command: %r", rest)
        sys.exit(1)

    validate_SNAP_module(args.module)

    log.info("Defaulting NV Params for module %s", args.module)

    _eprint_reset_message(args.module)

    eprint('Recovering Default NV Params:')
    widgets = [progressbar.FormatLabel('Processing'), ' ',
               progressbar.Percentage(), ' ', progressbar.Bar(), ' ',
               progressbar.Timer(), ' ', progressbar.AdaptiveETA()]

    progress = progressbar.ProgressBar(widgets=widgets, max_value=1000)

    def on_progress(counter, new_max):
        """Update Progress Bar"""
        progress.max_value = max(new_max, counter)
        progress.update(counter)

    def on_finished():
        """Finish Progress Bar"""
        progress.finish()

    progress.start()

    try:
        snapstraction.flash.default_nv(context.profile,
                                       args.module,
                                       progress_cb=on_progress,
                                       finished_cb=on_finished)
    except serialwrapper.PyserialDriver.SerialOpenException:
        exit_with_error("Could not open serial port: {0}".format(getattr(Profile.load('default'), 'device')))


def erase(parser, context, args):
    """Erase Script via recovery"""
    handler = ArgumentHandler(
        usage="snap [global options] recover erase-script [-h] <module>",
        description="Erase the script on the bridge node",
    )

    handler.add_argument('module',
                         help="Module name (i.e. 'RF220UF1', 'SS200')")

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to recover erase-script command: %r", rest)
        sys.exit(1)

    validate_SNAP_module(args.module)

    log.info("Erasing SNAPpy Image for module %s", args.module)

    _eprint_reset_message(args.module)

    eprint('Erase Script:')
    widgets = [progressbar.FormatLabel('Processing:'), ' ',
               progressbar.Percentage(), ' ', progressbar.Bar(), ' ',
               progressbar.Timer(), ' ', progressbar.AdaptiveETA()]

    progress = progressbar.ProgressBar(widgets=widgets, max_value=1000)

    def on_progress(counter, new_max):
        """Update Progress Bar"""
        progress.max_value = max(new_max, counter)
        progress.update(counter)

    def on_finished():
        """Finish Progress Bar"""
        progress.finish()

    progress.start()

    try:
        snapstraction.flash.erase_script(context.profile,
                                         args.module,
                                         progress_cb=on_progress,
                                         finished_cb=on_finished)
    except serialwrapper.PyserialDriver.SerialOpenException:
        exit_with_error("Could not open serial port: {0}".format(getattr(Profile.load('default'), 'device')))
    except snapstraction.flash.FlasherError:
        exit_with_error("Flasher Error: data timeout")
