import logging
import sys

from arghandler import ArgumentHandler

from tornado.ioloop import IOLoop, PeriodicCallback

from snaptoolbelt.util import eprint
from snapstraction.util import executing_in_virtualenv, get_virtualenv_path

log = logging.getLogger(__name__)


def script(parser, context, args):  # noqa - complexity
    """Portal-ish Script commands"""
    handler = ArgumentHandler(
        usage="snap [global options] script [-h] <script>",
        description="Run the specified SNAP (not SNAPpy!) script.",
    )

    handler.add_argument("script")

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to script command: %r", rest)
        eprint("Unexpected arguments to script command: %r", rest)
        sys.exit(1)

    startup_funcs = []
    on100ms_funcs = []
    on1s_funcs = []

    functions = {}

    def register_startup(func):
        startup_funcs.append(func)
        return func

    def register_100ms(func):
        on100ms_funcs.append(func)
        return func

    def register_1s(func):
        on1s_funcs.append(func)
        return func

    handlers = {}

    def register_handler(funcname):
        def reg_handler(function):
            handlers[funcname] = function
            return function
        return reg_handler

    gus = context.gus()

    def send_rpc(target, func, *args):
        IOLoop.current().add_callback(gus.ucast_rpc, target, func, *args)

    def send_mcast_rpc(group, ttl, func, *args):
        IOLoop.current().add_callback(gus.mcast_rpc, func, *args, group=group, ttl=ttl)

    def send_dmcast_rpc(targets, group, ttl, delay, func, *args):
        IOLoop.current().add_callback(gus.dmcast_rpc, targets, func, *args, group=group, ttl=ttl, delay=delay)

    script_globals = {
        'hook_startup': register_startup,
        'hook_100ms': register_100ms,
        'hook_1s': register_1s,
        'on': register_handler,
        'rpc': send_rpc,
        'mcastRpc': send_mcast_rpc,
        'mcast_rpc': send_mcast_rpc,
        'dmcastRpc': send_dmcast_rpc,
        'dmcast_rpc': send_dmcast_rpc
    }

    # Fix up the path to include wherever this script lives...
    if executing_in_virtualenv():
        sys.path.append(get_virtualenv_path())

    globals_before = script_globals.copy()

    execfile(args.script, script_globals, script_globals)

    log.debug("GLOBALS POST-EXEC: %r", script_globals)

    def dispatch_rpc(src, funcname, *args):
        log.info("RPC: %r.%r%r", src, funcname, args)

        func = handlers.get(funcname, None) or functions.get(funcname, None)
        if func:
            try:
                new_locals = {}
                new_locals['func'] = func
                new_locals['src'] = src
                new_locals['args'] = args
                exec "func(src, *args)" in script_globals, new_locals
            except:
                log.exception("Handling RPC: %r.%r%r", src, funcname, args)

    gus.on_any_rpc(dispatch_rpc)

    functions = {name: f for name, f in script_globals.iteritems() if callable(f) and name not in globals_before}

    log.info("FUNCTIONS: %r", functions)

    def mk_function_runner(funclist, name=None):
        name = name or "runner"

        def func_runner():
            for f in funclist:
                try:
                    new_locals = {}
                    new_locals['f'] = f
                    exec "f()" in script_globals, new_locals
                except:
                    log.exception(name)

        return func_runner

    run_startup_funcs = mk_function_runner(startup_funcs)
    run_100ms_funcs = mk_function_runner(on100ms_funcs)
    run_1s_funcs = mk_function_runner(on1s_funcs)

    try:
        gus.start()
        IOLoop.current().add_callback(run_startup_funcs)
        on100ms_pcb = PeriodicCallback(run_100ms_funcs, 100)
        on1s_pcb = PeriodicCallback(run_1s_funcs, 1000)
        on100ms_pcb.start()
        on1s_pcb.start()
        IOLoop.current().start()
    finally:
        gus.stop()
