import logging
import sys

from arghandler import ArgumentHandler

from tornado.ioloop import IOLoop

from snapstraction.broker import SNAPzmqBroker
from snapstraction.broker_client import get_active_profiles, get_pid_and_port_for_profile, kill_connection, broker_dir

from snaptoolbelt.util import eprint
from snapstraction.util import mkdir_p


log = logging.getLogger(__name__)


def start(parser, context, args):
    """Start connection sharing"""
    handler = ArgumentHandler(
        usage="snap [global options] start [-h]",
        description="Connect to the specified bridge and share the connection locally.",
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to start command: %r", rest)
        eprint("Unexpected arguments to start command: %r", rest)
        sys.exit(1)

    # Create broker dir
    try:
        mkdir_p(broker_dir)
    except:
        log.error("Couldn't create directory %s for connection sharing, you don't own your own home folder?", broker_dir)
        eprint("Couldn't create directory %s for connection sharing, you don't own your own home folder?", broker_dir)
        sys.exit(1)

    pid, port = get_pid_and_port_for_profile(context.profile)

    if port:
        log.info("Already running: pid %s on port %s", pid, port)
        context.output({"profile": context.profile.name, "port": port})
        sys.exit(0)

    broker = SNAPzmqBroker(context.profile, context.network, context.license)

    # Use a callback to print out the profile and port *after* we've started...
    def startup():
        context.output({"profile": context.profile.name, "pid": broker.pid, "port": broker.port})

    broker.on_startup(startup)
    broker.start()

    # Start serving
    try:
        IOLoop.current().start()
    finally:
        broker.stop()


def status(parser, context, args):
    """Show status of connection sharing"""
    handler = ArgumentHandler(
        usage="snap [global options] status [-h]",
        description="Show status of shared connections.",
    )

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to status command: %r", rest)
        eprint("Unexpected arguments to status command: %r", rest)
        sys.exit(1)

    for p in get_active_profiles():
        log.debug("profile: %r", p)
        pid, port = get_pid_and_port_for_profile(p)
        context.output({"profile": p.name, "pid": pid, "port": port})
    sys.exit(0)


def stop(parser, context, args):
    """Stop connection sharing"""

    handler = ArgumentHandler(
        usage="snap [global options] stop [-h] [all]",
        description="Command an existing shared bridge connection to stop sharing. (All stops all active sharing.)",
    )

    handler.add_argument('all', nargs='?', default=None)

    args, rest = handler.parse_known_args(args)

    if rest:
        log.error("Unexpected arguments to stop command: %r", rest)
        eprint("Unexpected arguments to stop command: %r", rest)
        sys.exit(1)

    if args.all and args.all.lower() == "all":
        for p in get_active_profiles():
            log.info("Disconnecting: %s", p.name)
            kill_connection(p)
    else:
        kill_connection(context.profile)
