"""SNAPtoolbelt utility functions."""

from __future__ import print_function

import errno
import os
import pip
import re
import sys
import traceback

from snap_modules import get_parts

from snapstraction.util import validate_snap_address


def mkdir_p(path):
    """Python equivalent of mkdir -p."""
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def eprint(*args, **kwargs):
    """Method to print the message on stderr instead of stdout."""
    print(*args, file=sys.stderr, **kwargs)


def render_error(err_msg, help_msg=None, *args, **kwargs):
    """Method to Uniformly render error message.

    :param err_msg: Main Error Message to display. Mandatory arg.
    :param help_msg: Extra help message for the user. Optional arg.
    """
    # If we've been handed an exception, use the first level of the traceback as the message.
    # This gives us something much better to work from if we're trying to figure out what
    # went wrong.
    if isinstance(err_msg, Exception):
        # ex = err_msg
        err_msg = traceback.format_exc(1)

    if help_msg:
        text = "{0}\n\n{1}".format(err_msg, help_msg)
    else:
        text = err_msg

    eprint(text)


def exit_with_error(err_msg, help_msg=None, exit_code=1, *args, **kwargs):
    """Render the provided error and exit."""
    render_error(err_msg, help_msg=help_msg, *args, **kwargs)
    sys.exit(exit_code)


def exit_with_usage(handler, log, *args, **kwargs):
    """Log, print usage, and exit"""
    exit_code = kwargs.get('exit_code', 1)
    log.error(*args, **kwargs)
    handler.print_help(sys.stderr)
    sys.exit(exit_code)


def log_and_exit(logger, err_msg, help_msg=None, exit_code=1, *args, **kwargs):
    """Log the error message to the logger, then exit with error."""
    logger.error(err_msg, *args, **kwargs)
    exit_with_error(err_msg, help_msg=help_msg, exit_code=exit_code, *args, **kwargs)


def validate_package(
        package_name, url="https://update.synapse-wireless.com/pypi",
        *args, **kwargs):
    """Validate package installed in the environment.

    :param  package_name: Name of the package that need to validate.
    :param url: Artifactory url to install package.
    """
    snap = 'snap'
    # String transformations done here to map package names to product names
    original_package_name = package_name
    if package_name not in get_installed_packages():
        if package_name.lower().startswith(snap):
            package_name = package_name.replace(snap, snap.upper())

        if 'firmware' in package_name:
            package_name = package_name.replace('-', '').replace('firmware', 'core ')

        pkg_msg = "{} is not installed!".format(package_name)
        resolve_msg = "Run `pip install -i {0} {1}` to fix this issue.".format(url, original_package_name)

        exit_with_error(pkg_msg, resolve_msg)


def get_installed_packages():
    """Method returns list of installed packages to mock in the testcase.

    :return list: list of packages installed in the current virtualenv
    """
    return [i.key for i in pip.get_installed_distributions()]


def validate_mac_address(target):
    """Method to validate proper mac address of a snap device.

    :param target: 6 digit hexadecimal snap address.
    """
    invalid_address = ''
    if isinstance(target, list) or isinstance(target, tuple):
        target = iter(target)
        for t in target:
            proper_addr = validate_snap_address(t)
            if not proper_addr:
                invalid_address = t
                break
    else:
        proper_addr = validate_snap_address(target)
        if not proper_addr:
            invalid_address = target

    if not proper_addr:
        exit_with_error("Invalid target address: {0}".format(invalid_address),
                        "Valid SNAP addresses are 6 hex digits, e.g. 'ABC123'..")


def key_error_with_options(message, options):
    """Formulate the input message and options and render the error.

    :param message: Message to be displayed on error.
    :param options: Takes List/Tuple/Function object and creates help text.
    """
    if isinstance(options, list) or isinstance(options, tuple):
        options = '\n'.join(options)
    else:
        options = '\n'.join(options())

    help_text = "Options are : \n%s" % options
    exit_with_error(message, help_text)


def get_config_network_keys():
    """Method to return keys list for network subcommand of config command.

    :returns keys_list: list of possible keys for network subcommand in config.
    """
    return [
        'channel', 'default_cb_retries', 'default_cb_timeout',
        'default_delay', 'default_group', 'default_ttl',
        'encryption_key', 'encryption_type', 'name', 'net_id',
        'packet_crc', 'second_crc']


def get_config_profile_keys():
    """Method to return keys list for profile subcommand of config command.

    :returns keys_list: list of possible keys for the profile subcommand in config.
    """
    return ['address', 'device', 'kind', 'name', 'network', 'speed']


def validate_SNAPcore_version(version):
    """Method to validate SNAPcore version.

    :param version: digit SNAPcore version.
    """
    version_regex = r"^\d+\.\d+\.\d+.*$"
    if not re.search(version_regex, version):
        exit_with_error("Invalid firmware: %s" % version, "Valid firmware version could be 2.8.2, 2.8.2rc1, 2.8.2a1, 2.8.2b1")


def validate_rpc_function_name(rpc_name):
    """Method to validate the function name is a valid python function name

    :param rpc_name: Name of the SNAPpy function
    """
    valid_rpc_name_regex = r"^[a-zA-Z]\w*$"
    if isinstance(rpc_name, basestring) and not re.search(valid_rpc_name_regex, rpc_name):
        exit_with_error("Invalid RPC function: {0}".format(rpc_name),
                        "Valid RPC functions should start with a character, "
                        "can contain '_' and can be alphanumeric. e.g. data_2_send, getData")


def validate_SNAP_module(module_name):
    """Validation of SNAP modules

    :param module_name: Name of the SNAP module(ex:RF100,SS200,..)
    """
    if module_name not in get_parts():
        key_error_with_options("Invalid module: %s" % module_name, get_parts())
